#!/bin/sh

# Poisson Media Recovery
# 
# The script is in public domain. It was originally written by
# Heikki Orsila <heikki.orsila@iki.fi> at 2005-06-20.

# This script tries to obtain a broken file from cdrom devices by repeatedly
# reading blocks from the media.
# It sequentially mounts and umounts the device, and on each mount it tries
# to read as many new 512 byte sectors as possible.
# 
# I was able to use this tool to save some files from old CDs.
# 
# USAGE:
#
# poisson_media_recovery.sh DEVICEMOUNT FULL_FILE_PATH RESULT_FILE
#
# DEVICEMOUNT (e.g. /mnt/cdrom) must be a valid mount path (see /etc/fstab).
# FULL_FILE_PATH (e.g. /mnt/cdrom/myfile) is an absolute path into the file
# to be recovered. RESULT_FILE (e.g. /tmp/myfile) is the path to the file
# that needed to be recovered.
#
# Note that /mnt/cdrom must not be mounted initially!
# 
# example:
# poisson_media_recovery.sh /mnt/cdrom /mnt/cdrom/mybrokenfile /tmp/savedfile

mountpath="$1"
filepath="$2"
resultfile="$3"

size="0"

function error {
    echo "fatal error:" "$@"
    exit -1
}

if test -z "$1" ; then
    error "read the script source code for instructions"
fi

if test -e "$resultfile" ; then error "$resultfile exists" ; fi

tmpfile="`tempfile`" || error "can not create temprorary file"
echo "using $tmpfile as a temporary file..."

while true ; do
    mount "$mountpath"
    if test "$?" != "0" ; then
	rm -f "$tmpfile"
	error "can not mount $mountpath"
    fi
    fullsize="`stat -c "%s" "$filepath"`"
    if test -z "$fullsize" ; then
	rm -f "$tmpfile"
	error "can not stat $filepath"
    fi
    echo "starting reading from offset $size"
    dd if="$filepath" bs=512 skip=$((size/512)) > "$tmpfile"
    retval="$?"
    umount "$mountpath"
    if test "$retval" = "0" ; then
	cat "$tmpfile" >> "$resultfile"
	break
    else
	echo "an error detected"
	newsize="`stat -c "%s" "$tmpfile"`"
	if test -z "$newsize" ; then
	    rm -f "$tmpfile"
	    error "can not stat $tmpfile"
	fi
	addition=$((newsize/512))
	if test "$addition" -gt "0" ; then
	    echo "addition of $addition * 512 bytes"
	    size=$((size + addition * 512))
	    echo "new recovered size is $size bytes"
	    dd if="$tmpfile" bs=512 count="$addition" >> "$resultfile"
	else
	    # this code path has never been tested :-)
	    remainder=$((fullsize - size))
	    if "$newsize" -eq "$remainder" ; then
		cat "$tmpfile" >> "$resultfile"
		break
	    fi
	    echo "no new data blocks. retrying.."
	fi
    fi
done

rm -f "$tmpfile"
